#!/usr/bin/env python3

# -*- coding: UTF-8 -*-

"""
Description: The goal of this script is to check whether communities \
are often regulated by a splicing factor
"""

from typing import List, Tuple, Dict
import sqlite3
from .config import ConfigGraph
from ..figures_utils.exons_interactions import get_every_events_4_a_sl
from .community_finder import get_communities
import pandas as pd
import numpy as np
from itertools import product
from .community_finder import get_projects_name
import multiprocessing as mp
import logging
from ..logging_conf import logging_def
import doctest
from functools import reduce
from operator import add
import statsmodels.formula.api as api
from statsmodels.stats.multitest import multipletests
from rpy2.robjects import r, pandas2ri
from pathlib import Path


def get_exon_regulated_in_communities(community: List[str],
                                      regulated_exons: List[str]
                                      ) -> Tuple[int, int, float]:
    """

    :param community: strongly co-localized group of exons belonging to \
    the same community.
    :param regulated_exons: List of regulated exons by a splicing factor
    :return: The number of exons regulated by the splicing factor inside \
    the community. The total number of exons within the community and \
    the percentage of regulated exons within the community
    """
    reg_community = len([e for e in community if e in regulated_exons])
    tot_commmunity = len(community)
    prop = reg_community / tot_commmunity * 100
    return reg_community, tot_commmunity, prop


def get_ctrl_exon(cnx: sqlite3.Connection) -> List[str]:
    """
    Get the list of fasterDB exons.

    :param cnx: Connection to chIA-PET database
    :return: The list of fasterDB exons
    """
    cursor = cnx.cursor()
    query = 'SELECT id FROM cin_exon'
    cursor.execute(query)
    res = np.asarray(cursor.fetchall()).flatten()
    cursor.close()
    return list(res)


def get_stat_community_exons(cnx: sqlite3.Connection, dic: Dict[str, List],
                             communities: List[List[str]],
                             regulated_exons: List[str]):
    """
    Get the data for the list of exons belonging to a community, and update \
    `dic`with it.

    :param cnx: connection to ChIA-PET database
    :param dic: Dictionary containing data about a community and it's content \
    corresponding to eoxns regulated by a splicing factor.
    :param communities: strongly co-localized group of exons belonging to \
    the same community.
    :param regulated_exons: List of regulated exons by a splicing factor
    :return: The update dictionary and the number of regulated exon by a \
    splicing factor that belong to a large community and the number of \
    exons belonging to a large community.
    """
    full_com = reduce(add, communities)
    ctrl_exons = get_ctrl_exon(cnx)
    n_reg, n_tot = np.nan, np.nan
    for cur_com, name in [[full_com, 'All-community'], [ctrl_exons, "FASTERDB"]]:
        reg_community, tot_commmunity, prop = \
            get_exon_regulated_in_communities(cur_com, regulated_exons)
        if name == 'All-community':
            n_reg = reg_community
            n_tot = tot_commmunity
        dic["community"].append(name)
        dic['reg_in_community'].append(reg_community)
        dic['community_size'].append(tot_commmunity)
        dic['%reg in community'].append(prop)
        dic['pval'].append(np.nan)
    return dic, n_reg, n_tot


def get_vector(n: int, t: int) -> List[int]:
    """
    Get a vector containing n ones and t - n zeros.

    :param n: The number of ones wanted
    :param t: The size of the array
    :return: The vector of n one's and t zeros

    >>> get_vector(2, 5)
    [1, 1, 0, 0, 0]
    """
    return [1] * n + [0] * (t - n)


def logistic_regression(n1: int, t1: int, n2: int, t2: int) -> float:
    """
    Get the p-value of the logisitcal test performed on the \
    test vector data defined by n1 ones and t1-n1 zeros and the control \
    vector data defined by n2 ones and t2-n2 zeros.

    :param n1: number of ones in test vectors
    :param t1: size of test vector
    :param n2: number of ones in control vector
    :param t2: size of control vector
    :return: The p-value of the logistic test
    """
    if n1 == 0 or n2 == 0 or n1 == t1 or n2 == t2:
        return 1
    df = pd.DataFrame({'y': get_vector(n1, t1) +
                            get_vector(n2, t2),
                       'x': ['test'] * t1 + ['ctrl'] * t2})
    return api.logit('y ~ x', data=df).fit(disp=False).pvalues['x[T.test]']


def expand_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    """
    Expend the dataframe df to be able to make a general mixed model on it.

    :param df: The dataframe to expand
    :return:  The dataframe expanded

    >>> d = pd.DataFrame({"community": ["C0", "C1"],
    ... "reg_in_community": [2, 3],
    ... "community_size": [5, 7],
    ... "%reg in community": [40, 42.85], 'pval': [1, 0.5], 'padj': [1, 1]})
    >>> expand_dataframe(d)
        is_reg community  reg_in_community  community_size
    0        1        C0                 2               5
    1        1        C0                 2               5
    2        0        C0                 2               5
    3        0        C0                 2               5
    4        0        C0                 2               5
    5        1        C1                 3               7
    6        1        C1                 3               7
    7        1        C1                 3               7
    8        0        C1                 3               7
    9        0        C1                 3               7
    10       0        C1                 3               7
    11       0        C1                 3               7

    """
    content = []
    for row in df.values:
        row = list(row)
        values = [1] * row[1] + [0] * (row[2] - row[1])
        for v in values:
            content.append([v] + row)
    df = pd.DataFrame(content, columns=["is_reg"] + list(df.columns))
    df.drop(["%reg in community", "pval", "padj"], axis=1, inplace=True)
    return df


def glmm_maker(expanded_df: pd.DataFrame, outfile: Path) -> float:
    """
    Make the glmm analysis to see if the exon regulated by a splicing factor \
    are equally distributed among the communities.

    :param expanded_df: The expanded dataframe
    :param outfile: A name of a file
    :return: the pvalue of glmm

    >>> d = pd.DataFrame({"community": ["C0", "C1"],
    ... "reg_in_community": [2, 3],
    ... "community_size": [5, 7],
    ... "%reg in community": [40, 42.85], 'pval': [1, 0.5], 'padj': [1, 1]})
    >>> e_df = expand_dataframe(d)
    >>> outfile = ConfigGraph.get_community_file("Test", 1, 1, True,
    ... "_stat.txt", True)
    >>> glmm_maker(e_df, outfile)
    1.0
    """
    pandas2ri.activate()
    glmm = r(
        """
        require("lme4")
        require("DHARMa")
        
        function(data, folder, partial_name) {
            null_mod <- glm(is_reg ~ log(community_size) , family=binomial, data=data)
            mod <- glmer(is_reg ~ log(community_size) + (1 | community), data=data, family=binomial, nAGQ=0)
            simulationOutput <- simulateResiduals(fittedModel = mod, n = 250)
            png(paste0(folder, "/dignostics_", partial_name, ".png"))
            plot(simulationOutput)
            dev.off()
            return(anova(mod, null_mod, test="Chisq"))
        }
        
        """)
    folder = outfile.parent / "diagnostics"
    folder.mkdir(parents=True, exist_ok=True)
    partial_name = outfile.name.replace('.txt', '')
    return glmm(expanded_df, str(folder), partial_name)["Pr(>Chisq)"][1]


def glmm_statistics(df: pd.DataFrame, sf_name: str, reg: str,
                    project: str, weight: int, global_weight: int,
                    same_gene: bool) -> pd.Series:
    """
    Create the glmm statistics for a given splicing factor with \
    given communities.

    :param df: dataframe contaning stats data.
    :param sf_name: The splicing factor of interest
    :param reg: The regulation of interest
    :param project: The name of the project of interest
    :param weight: The minimum weight of interaction to consider
    :param global_weight: The global weight to consider. if \
    the global weight is equal to 0 then then density figure are calculated \
    by project, else all projet are merge together and the interaction \
    seen in `global_weight` project are taken into account
    :param same_gene: Say if we consider as co-localised, exons within the \
    same gene (True) or not (False) (default False)
    :return: The glmm pvalue among with other informations
    """
    ndf = df.loc[-df['community'].isin(["All-community", "FASTERDB"]),
          :].copy()
    expanded_df = expand_dataframe(ndf)
    outfile = ConfigGraph.get_community_file(project, weight, global_weight,
                                             same_gene,
                                             f"{sf_name}_{reg}_stat.txt", True)
    noutfold = outfile.parent / "expanded_df"
    noutfold.mkdir(exist_ok=True, parents=True)
    noutfile = noutfold / outfile.name
    expanded_df.to_csv(noutfile, sep="\t", index=False)
    pval = glmm_maker(expanded_df, outfile)
    return pd.Series({"project": project, "sf": sf_name, "reg": reg,
                      "pval": pval})


def get_stat4communities(sf_name: str, reg: str,
                         project: str, weight: int,
                         global_weight: int, same_gene: bool
                         ) -> Tuple[pd.DataFrame, pd.Series]:
    """
    Get data (proportion of `reg` regulated exons by a splicing factor
    `sf_name` within a community) for every community.

    :param sf_name: The splicing factor of interest
    :param reg: The regulation of interest
    :param project: The name of the project of interest
    :param weight: The minimum weight of interaction to consider
    :param global_weight: The global weight to consider. if \
    the global weight is equal to 0 then then density figure are calculated \
    by project, else all projet are merge together and the interaction \
    seen in `global_weight` project are taken into account
    :param same_gene: Say if we consider as co-localised, exons within the \
    same gene (True) or not (False) (default False)
    """
    logging.debug(f"Working on {sf_name}-{reg}, for {project}, w:{weight}, "
                  f"g:{global_weight}")
    cnx = sqlite3.connect(ConfigGraph.db_file)
    result = ConfigGraph.get_community_file(project, weight, global_weight,
                                            same_gene,
                                            "_communities.txt")
    communities = get_communities(result, 0)
    regulated_dic, number = get_every_events_4_a_sl(cnx, sf_name, reg)
    reg_exons = regulated_dic[sf_name + "_" + reg]
    d = {"community": [], "reg_in_community": [], "community_size": [],
         "%reg in community": [], 'pval': []}
    d, n2, t2 = get_stat_community_exons(cnx, d, communities, reg_exons)
    for k, community in enumerate(communities):
        if len(community) > 49:
            reg_community, tot_commmunity, prop = \
                get_exon_regulated_in_communities(community, reg_exons)
            d["community"].append(f"C{k + 1}")
            d['reg_in_community'].append(reg_community)
            d['community_size'].append(tot_commmunity)
            d['%reg in community'].append(prop)
            d['pval'].append(logistic_regression(reg_community, tot_commmunity,
                                                 n2, t2))
    d['padj'] = [np.nan, np.nan] + list(multipletests(d['pval'][2:],
                                                      method='fdr_bh')[1])
    d['sf'] = [sf_name] * len(d["community"])
    d['reg'] = [reg] * len(d["community"])
    d['project'] = [project] * len(d["community"])
    df = pd.DataFrame(d)
    s = glmm_statistics(df, sf_name, reg, project, weight, global_weight,
                        same_gene)
    return df, s


def get_sfname() -> List[str]:
    """
    Recover the name of every splicing factor of interest.

    :return: The list of splicing factor

    >>> len(get_sfname()) == 42
    True
    """
    logging.debug('recovering sf factors ...')
    return ["PCBP2", "HNRNPA1", "HNRNPU", "QKI", "PTBP1", "TRA2A_B", "KHSRP",
            "MBNL1", "HNRNPL", "HNRNPK", "SRSF7", "HNRNPA2B1", "SFPQ", "RBM15",
            "HNRNPM", "FUS", "DAZAP1", "RBM39", "SRSF9", "RBM25", "RBM22",
            "HNRNPF", "SRSF5", "PCBP1", "RBFOX2", "HNRNPH1", "RBMX", "SRSF6",
            "MBNL2", "SRSF1","SRSF2", "HNRNPC", "SRSF3", "U2AF1", "SF3B1",
            "SNRNP70", "SNRPC", "DDX5_DDX17", "SF1", "SF3A3", "SF3B4", "U2AF2"]


def get_key(project: str, weight: int) -> str:
    """
    Get the combination of project and weight parameters

    :param project: A project name
    :param weight: A weight
    :return: The string combination of the two
    """
    return f"{project}_{weight}"


def multiple_stat_launcher(ps: int, weights: List[int],
                           global_weights: List[int],
                           same_gene: bool, logging_level: str = "DISABLE"):
    """
    Launch the statistical analysis for every

    :param ps: The number of processes we want to use.
    :param weights: The list of weights of interaction to consider
    :param global_weights: The list global weights to consider. if \
    the global weight is equal to 0 then then density figure are calculated \
    by project, else all projcet are merge together and the interaction \
    seen in `global_weight` project are taken into account
    :param same_gene: Say if we consider as co-localised exon within the \
    same gene
    :param logging_level: Level of information to display
    """
    ConfigGraph.community_folder.mkdir(exist_ok=True, parents=True)
    logging_def(ConfigGraph.community_folder, __file__, logging_level)
    logging.info("Checking if communities contains often exons regulated by a"
                 " splicing factor")
    sf_list = get_sfname()
    global_weights = list(np.unique(global_weights))
    weights = list(np.unique(weights))
    projects, dic_project = get_projects_name(global_weights)
    condition = list(product(projects, weights, sf_list, ['down', 'up']))
    processes = {}
    pool = mp.Pool(processes=min(ps, len(condition)))
    logging.debug("Calculating stats...")
    for project, weight, sf_name, reg in condition:
        ckey = get_key(project, weight)
        global_weight = dic_project[project]
        args = [sf_name, reg, project, weight, global_weight, same_gene]
        if ckey not in processes.keys():
            processes[ckey] = [pool.apply_async(get_stat4communities, args)]
        else:
            processes[ckey].append(pool.apply_async(get_stat4communities, args))
    for p, value in processes.items():
        project, weight = p.split("_")
        list_tuples = [proc.get(timeout=None) for proc in value]
        pool.close()
        pool.join()
        list_df = [t[0] for t in list_tuples]
        list_series = [t[1] for t in list_tuples]
        df = pd.concat(list_df, axis=0, ignore_index=True)
        outfile = ConfigGraph.get_community_file(project, weight,
                                                 dic_project[project],
                                                 same_gene,
                                                 "_stat.txt", True)
        df.to_csv(outfile, sep="\t", index=False)
        glm_df = pd.DataFrame(list_series)
        glm_df["padj"] = multipletests(glm_df['pval'].values,
                                       method='fdr_bh')[1]
        outfile = ConfigGraph.get_community_file(project, weight,
                                                 dic_project[project],
                                                 same_gene,
                                                 "_glmm_stat.txt", True)
        glm_df.to_csv(outfile, sep="\t", index=False)


if __name__ == "__main__":
    doctest.testmod()